overlay coq_dpdgraph https://github.com/SkySkimmer/coq-dpdgraph dirpath-full-path 20371

overlay coq_lsp https://github.com/SkySkimmer/coq-lsp dirpath-full-path 20371

overlay coqhammer https://github.com/SkySkimmer/coqhammer dirpath-full-path 20371

overlay quickchick https://github.com/SkySkimmer/QuickChick dirpath-full-path 20371

overlay tactician https://github.com/SkySkimmer/coq-tactician dirpath-full-path 20371

overlay vscoq https://github.com/SkySkimmer/vscoq dirpath-full-path 20371
