overlay coq_lsp https://github.com/ppedrot/coq-lsp module-abstract-type 19995

overlay elpi https://github.com/ppedrot/coq-elpi module-abstract-type 19995

overlay metacoq https://github.com/ppedrot/metacoq module-abstract-type 19995

overlay paramcoq https://github.com/ppedrot/paramcoq module-abstract-type 19995
