overlay coq_lsp https://github.com/SkySkimmer/coq-lsp def-loc 20272

overlay equations https://github.com/SkySkimmer/Coq-Equations def-loc 20272

overlay elpi https://github.com/SkySkimmer/coq-elpi def-loc 20272

overlay quickchick https://github.com/SkySkimmer/QuickChick def-loc 20272
