overlay elpi https://github.com/SkySkimmer/coq-elpi template-entry-qvar 20178

overlay equations https://github.com/SkySkimmer/Coq-Equations template-entry-qvar 20178

overlay coq_lsp https://github.com/SkySkimmer/coq-lsp template-entry-qvar 20178

overlay metacoq https://github.com/SkySkimmer/metacoq template-entry-qvar 20178

overlay paramcoq https://github.com/SkySkimmer/paramcoq template-entry-qvar 20178
