overlay elpi https://github.com/ppedrot/coq-elpi abstract-clean-seff 20472

overlay itauto https://gitlab.inria.fr/pedrot/itauto abstract-clean-seff 20472

overlay waterproof https://github.com/ppedrot/coq-waterproof abstract-clean-seff 20472
