overlay elpi https://github.com/proux01/coq-elpi stdlib_repo 19530
overlay metacoq https://github.com/proux01/metacoq stdlib_repo 19530
overlay quickchick https://github.com/proux01/QuickChick stdlib_repo 19530

overlay coqutil https://github.com/proux01/coqutil stdlib_repo 19530
overlay coq_dpdgraph https://github.com/coq-community/coq-dpdgraph stdlib_repo 19530
overlay equations https://github.com/proux01/Coq-Equations stdlib_repo 19530
overlay coq_tools https://github.com/proux01/coq-tools stdlib_repo 19530
overlay coq_lsp https://github.com/proux01/coq-lsp stdlib_repo 19530
