overlay elpi https://github.com/SkySkimmer/coq-elpi simpl-proj-flags 20605

overlay metarocq https://github.com/SkySkimmer/metarocq simpl-proj-flags 20605

overlay lean_importer https://github.com/SkySkimmer/rocq-lean-import simpl-proj-flags 20605

overlay coq_lsp https://github.com/SkySkimmer/coq-lsp simpl-proj-flags 20605
