overlay equations https://github.com/ppedrot/Coq-Equations namegen-no-global 20423

overlay lean_importer https://github.com/ppedrot/rocq-lean-import namegen-no-global 20423

overlay rewriter https://github.com/ppedrot/rewriter namegen-no-global 20423
