overlay stdlib https://github.com/jrosain/rocq-stdlib more-uniform-elimination 20397

overlay coq_lsp https://github.com/jrosain/rocq-lsp more-uniform-elimination 20397

overlay elpi https://github.com/jrosain/rocq-elpi more-uniform-elimination 20397

overlay equations https://github.com/jrosain/Rocq-Equations more-uniform-elimination 20397

overlay unicoq https://github.com/jrosain/unicoq more-uniform-elimination 20397

overlay tactician https://github.com/jrosain/rocq-tactician more-uniform-elimination 20397

overlay coqhammer https://github.com/jrosain/rocqhammer more-uniform-elimination 20397

overlay metacoq https://github.com/jrosain/metarocq more-uniform-elimination 20397
