Attributes deprecated(note="This library is useless.", since="XX YY").
Attributes warn(note="This library is dangerous.", cats="dangerous library").
Attributes warn(note="This library is tricky.", cats="dangerous library, tricky library").
