overlay paramcoq https://github.com/SkySkimmer/paramcoq cominductive-lbound 20102 kernel-template-qvar

overlay coq_lsp https://github.com/SkySkimmer/coq-lsp cominductive-lbound 20102 kernel-template-qvar

overlay metacoq https://github.com/SkySkimmer/metacoq kernel-template-qvar 20102
