overlay neural_net_interp https://github.com/SkySkimmer/neural-net-coq-interp ltacX-obligations 19690

overlay rupicola https://github.com/SkySkimmer/rupicola ltacX-obligations 19690
# Make PRs against https://github.com/mit-plv/rupicola base branch master

overlay mtac2 https://github.com/SkySkimmer/Mtac2 ltacX-obligations 19690

overlay paramcoq https://github.com/SkySkimmer/paramcoq ltacX-obligations 19690

overlay waterproof https://github.com/SkySkimmer/coq-waterproof ltacX-obligations 19690

overlay tactician https://github.com/SkySkimmer/coq-tactician ltacX-obligations 19690

overlay coq_lsp https://github.com/SkySkimmer/coq-lsp ltacX-obligations 19690

overlay rewriter https://github.com/SkySkimmer/rewriter ltacX-obligations 19690

overlay fiat_crypto https://github.com/SkySkimmer/fiat-crypto ltacX-obligations 19690
