overlay coq_lsp https://github.com/LeoAlexElouan/coq-lsp clarify-bound-names 20496
overlay lean_importer https://github.com/LeoAlexElouan/rocq-lean-import clarify-bound-names 20496
overlay metarocq https://github.com/LeoAlexElouan/metarocq clarify-bound-names 20496
overlay paramcoq https://github.com/LeoAlexElouan/paramcoq clarify-bound-names 20496
